#include "device.h"
#include "common\binary.h"
#include "common\command.h"
#include "common\convert.h"
#include "constant\constant.h"

using namespace System;
using namespace System::IO;
using namespace System::Text;
using namespace System::Collections::Generic;


namespace TechnoAP { namespace usbmca {
	
	using namespace FTD2XX_NET;
	using namespace TechnoAP::usbmca;

	namespace device
	{
		/// <summary>
		/// static constractor
		/// </summary>
		static device::device(void)
		{
			// Adding environment variables(PATH), Folder with driver DLL(ftd2xx.dll)
			String ^strOldPath = Environment::GetEnvironmentVariable("PATH", EnvironmentVariableTarget::Process);
			String ^strSystemPath = Environment::GetFolderPath(Environment::SpecialFolder::System);
			String ^strDriverPath = strOldPath + ";" + strSystemPath + "\\TechnoAP";
			Environment::SetEnvironmentVariable("PATH", strDriverPath);
		}

		/// <summary>
		/// Write command message
		/// </summary>
		bool device::WriteCommand(String ^command, int channel, unsigned int value)
		{
			FT_STATUS ftResult;
			bool result = false;

			// Generation of command telegram
			array<Byte> ^abCommand = common::command::createTelegram(command, channel, value);

			// Write command
			unsigned int nBytesWritten = 0;
			ftResult = this->Write(abCommand, abCommand->Length, nBytesWritten);

			mLastWriteCommandValue = value;
			mLastStatus = ftResult;

			if (ftResult == FT_STATUS::FT_OK && nBytesWritten == abCommand->Length) {
				result = true;
			}
			
			return	result;
		}

		/// <summary>
		/// Read command response message
		/// </summary>
		bool device::ReadCommand(String ^%command, unsigned int %value)
		{
			FT_STATUS ftResult;
			bool result = false;

			array<Byte> ^abCommand = gcnew array<Byte>(8);

			unsigned int nBytesRead = 0;
			ftResult = this->Read(abCommand, (unsigned int)(abCommand->Length), nBytesRead);

			if (ftResult == FT_STATUS::FT_OK && nBytesRead == abCommand->Length) {
				command = common::command::restoreCommand(abCommand, value);
				mLastReadCommandValue = value;
				result = true;
			}

			return	result;
		}

		/// <summary>
		/// Command execution
		/// </summary>
		bool device::OrderCommand(String ^command, int channel, unsigned int input, unsigned int %output)
		{
			bool result = false;
			bool bln1st, bln2nd;

			// Write request command
			bln1st = WriteCommand(command, channel, input);

			if (bln1st == true) {
				// Read response command
				String ^dummy = String::Empty;
				bln2nd = ReadCommand(dummy, output);
				result = bln2nd;
			}

			return	result;
		}

		/// <summary>
		/// Start measurement
		/// </summary>
		bool device::StartMeasurement(void)
		{
			return OrderCommand(const_cast<String^>(constant::Command::AQSW), (int)constant::CH_TYPE::Common, 1u, mOutputDummy);
		}

		/// <summary>
		/// Stop measurement
		/// </summary>
		bool device::StopMeasurement(void)
		{
			return OrderCommand(const_cast<String^>(constant::Command::AQEW), (int)constant::CH_TYPE::Common, 1u, mOutputDummy);
		}

		/// <summary>
		/// Clear measurement data
		/// </summary>
		bool device::ClearMeasurementData(void)
		{
			return OrderCommand(const_cast<String^>(constant::Command::CLRW), (int)constant::CH_TYPE::Common, 1u, mOutputDummy);
		}

		/// <summary>
		/// Setting measurement mode(historam or list)
		/// </summary>
		bool device::SetMeasurementDataMode(constant::DataMode value)
		{
			return OrderCommand(const_cast<String^>(constant::Command::MODW), (int)constant::CH_TYPE::Common, (UInt32)value, mOutputDummy);
		}

		/// <summary>
		/// Setting measurement time mode(real or live)
		/// </summary>
		bool device::SetMeasurementTimeMode(constant::TimeMode value)
		{
			return OrderCommand(const_cast<String^>(constant::Command::MMDW), (int)constant::CH_TYPE::Common, (UInt32)value, mOutputDummy);
		}

		/// <summary>
		/// Setting measurement time
		/// </summary>
		bool device::SetMeasurementTime(__int64 value)
		{
			bool result = false;

			UInt32 value0 = common::binary::getMsb32bit(value) & 0x0fff;
			bool bln0th = OrderCommand(const_cast<String^>(constant::Command::MT0W), (int)constant::CH_TYPE::Common, value0, mOutputDummy);

			if (bln0th == true) {
				UInt32 value1 = common::binary::getLsb32bit(value);
				bool bln1st = OrderCommand(const_cast<String^>(constant::Command::MT1W), (int)constant::CH_TYPE::Common, value1, mOutputDummy);
				result = bln1st;
			}

			return	result;
		}

		/// <summary>
		/// Setting of measurement time (in seconds)
		/// </summary>
		bool device::SetMeasurementTimeSeconds(__int64 value)
		{
			__int64 qwCount = common::convert::SecondsToCount(value);
			return SetMeasurementTime(qwCount);
		}

		/// <summary>
		/// Setting of measurement time (in milliseconds)
		/// </summary>
		bool device::SetMeasurementTimeMiilliseconds(__int64 value)
		{
			__int64 qwCount = common::convert::MillisecondsToCount(value);
			return SetMeasurementTime(qwCount);
		}

		/// <summary>
		/// Reading measurement status
		/// </summary>
		status^ device::ReadStatus(void)
		{
			status ^resut = nullptr;

			// Measurement state request
			bool blnResult = WriteCommand(const_cast<String^>(constant::Command::STUW), (int)constant::CH_TYPE::Common, 0);

			if (blnResult == true) {
				// Reading measurement status(response data)
				array<Byte> ^abStatus = gcnew array<Byte>(94);
				unsigned int uBytesRead = 0;
				FT_STATUS ftResult = Read(abStatus, (UInt32)abStatus->Length, uBytesRead);

				if (ftResult == FT_STATUS::FT_OK && abStatus->Length == (int)(uBytesRead)) {
					// Organization of measurement state data
					mStatus->analyzeData(abStatus);
					resut = mStatus;
				}
			}

			return resut;
		}

		/// <summary>
		/// Read list data
		/// </summary>
		array<Byte>^ device::ReadListData(int %nReadBytes)
		{
			bool bln1st, bln2nd;
			array<Byte> ^abListData = nullptr;

			nReadBytes = 0;

			// Request list data
			bln1st = WriteCommand(const_cast<String^>(constant::Command::LISR), (int)constant::CH_TYPE::Common, 0);

			if (bln1st == true) {
				// Read response command
				unsigned int uOutput = 0;
				String ^strCommand = String::Empty;
				bln2nd = ReadCommand(strCommand, uOutput);

				if (bln2nd == true) {
					if (strCommand->CompareTo(const_cast<String^>(constant::Command::LISA)) == 0) {
						int nBlock = (int)(uOutput) & 0x0000ffff;
						int nBytes = (int)(nBlock * (UInt32)constant::Size::ListDataBlock);

						if (nBytes > 0) {
							// There is list data -> Reading data
							unsigned int nReadBytesWork = 0;
							abListData = gcnew array<Byte>(nBytes);
							this->Read(abListData, (unsigned int)(nBytes), nReadBytesWork);
							nReadBytes = (int)(nReadBytesWork);
						}
					}
					else {
						// Invalid response command
						mnErrorReadList = -3;
					}
				}
				else {
					// Failed to read response command
					mnErrorReadList = -2;
				}
			}
			else {
				// Command write failure
				mnErrorReadList = -1;
			}

			return	abListData;
		}

		/// <summary>
		/// Reading histogram data
		/// </summary>
		/// <param name="channel">target CH</param>
		/// <param name="maxch">Max ch count(16384, 8192, 4096~)</param>
		/// <returns>command text</returns>
		array<unsigned int>^ device::ReadHistogram(int channel, int maxch)
		{
			array<unsigned int> ^auHistogram = gcnew array<unsigned int>(constant::Size::HistogramMaxCh);
			bool bln1st, bln2nd;
			unsigned int uInput, uOutput;

			Array::Clear(auHistogram, 0, auHistogram->Length);

			// Target CH switching
			uInput = (unsigned int)(channel - 1);
			uOutput = 0;
			bln1st = OrderCommand(const_cast<String^>(constant::Command::HCHW), (int)constant::CH_TYPE::Common, uInput, uOutput);

			if (bln1st == true) {
				unsigned int uRxBytes = 0;
				array<Byte> ^abRxData = gcnew array<Byte>(constant::Size::HistogramOnceByte);
				Array::Clear(abRxData, 0, abRxData->Length);

				int nRxCount = maxch / constant::Size::HistogramOnceCh;

				for (int i = 0; i < nRxCount; i++) {
					// Request the histogram data of the block
					String ^command = String::Format(const_cast<String^>(constant::Command::HIXX), i);
					bln2nd = WriteCommand(command, (int)constant::CH_TYPE::Common, 0);

					if (bln2nd == false) {
						auHistogram = nullptr;
						break;
					}

					// Read out the histogram data of this block
					unsigned int uBytesRead = 0;
					FT_STATUS ftResult = this->Read(abRxData, abRxData->Length, uBytesRead);

					if (ftResult != FT_STATUS::FT_OK || uBytesRead != abRxData->Length) {
						mLastStatus = ftResult;
						auHistogram = nullptr;
						break;
					}

					Buffer::BlockCopy(abRxData, 0, auHistogram, (int)uRxBytes, (int)uBytesRead);
					uRxBytes += uBytesRead;
				}

				// Convert to host byte order
				for (int i = 0; i < auHistogram->Length; i++) {
					int nData = (int)auHistogram[i];
					auHistogram[i] = (unsigned int)System::Net::IPAddress::NetworkToHostOrder(nData);
				}
			}

			return	auHistogram;
		}
	}
} }


